from .async_connectable import AsyncClosable, AsyncConnectable, AsyncConnectionManager, AsyncResourceManager
from .sync_connectable import Closable, Connectable, ConnectionManager, ResourceManager

__all__ = [
    "Closable", "Connectable", "ResourceManager", "ConnectionManager",
    "AsyncClosable", "AsyncConnectable", "AsyncResourceManager", "AsyncConnectionManager",
]



import abc
import asyncio
from typing import Self, override


class AsyncClosable(abc.ABC):

    @abc.abstractmethod
    async def close(self):
        pass

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class AsyncConnectable(abc.ABC):

    @abc.abstractmethod
    async def connect(self):
        pass

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return None


class AsyncResourceManager(AsyncConnectable, AsyncClosable, abc.ABC):

    @override
    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    @override
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class AsyncConnectionManager[T](AsyncResourceManager, abc.ABC):

    def __init__(self):
        super().__init__()
        self._connection: T | None = None
        self._lock = asyncio.Lock()

    @property
    def connected(self) -> bool:
        return self._connection is not None

    @override
    async def connect(self):
        async with self._lock:
            if self._connection is not None:
                raise ConnectionError('Connection is already created')
            self._connection = await self._create_connection()

    @override
    async def close(self):
        if self._connection is None:
            return
        async with self._lock:
            self._connection = None

    @property
    def _connected_conn(self) -> T:
        if self._connection is None:
            raise ConnectionError("Connection is not created")
        return self._connection

    @abc.abstractmethod
    async def _create_connection(self) -> T:
        pass



import abc
from typing import Self, override


class Closable(abc.ABC):

    @abc.abstractmethod
    def close(self):
        pass

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class Connectable(abc.ABC):

    @abc.abstractmethod
    def connect(self):
        pass

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return None


class ResourceManager(Connectable, Closable, abc.ABC):

    @override
    def __enter__(self) -> Self:
        self.connect()
        return self

    @override
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class ConnectionManager[T](ResourceManager, abc.ABC):

    def __init__(self):
        super().__init__()
        self._connection: T | None = None

    @property
    def connected(self) -> bool:
        return self._connection is not None

    @override
    def connect(self):
        if self._connection is not None:
            raise ConnectionError('Connection is already created')
        self._connection = self._create_connection()

    @override
    def close(self):
        self._connection = None

    @property
    def _connected_conn(self) -> T:
        if self._connection is None:
            raise ConnectionError("Connection is not created")
        return self._connection

    @abc.abstractmethod
    def _create_connection(self) -> T:
        pass
